﻿Imports Microsoft.VisualBasic
Imports System
Imports System.Collections.Generic
Namespace KMeansPlus
    Friend Class KMeansPlusProgram
        Shared Sub Main(ByVal args() As String)
            Console.WriteLine(vbCrLf & "Begin k-means++ clustering demo" & vbCrLf)
            ' real data likely to come from a text file or SQL
            Dim rawData(19)() As Double
            rawData(0) = New Double() {65.0, 220.0}
            rawData(1) = New Double() {73.0, 160.0}
            rawData(2) = New Double() {59.0, 110.0}
            rawData(3) = New Double() {61.0, 120.0}
            rawData(4) = New Double() {75.0, 150.0}
            rawData(5) = New Double() {67.0, 240.0}
            rawData(6) = New Double() {68.0, 230.0}
            rawData(7) = New Double() {70.0, 220.0}
            rawData(8) = New Double() {62.0, 130.0}
            rawData(9) = New Double() {66.0, 210.0}
            rawData(10) = New Double() {77.0, 190.0}
            rawData(11) = New Double() {75.0, 180.0}
            rawData(12) = New Double() {74.0, 170.0}
            rawData(13) = New Double() {70.0, 210.0}
            rawData(14) = New Double() {61.0, 110.0}
            rawData(15) = New Double() {58.0, 100.0}
            rawData(16) = New Double() {66.0, 230.0}
            rawData(17) = New Double() {59.0, 120.0}
            rawData(18) = New Double() {68.0, 210.0}
            rawData(19) = New Double() {61.0, 130.0}

            Console.WriteLine("Raw unclustered data:" & vbCrLf)
            Console.WriteLine("    Height Weight")
            Console.WriteLine("-------------------")
            ShowData(rawData, 1, True, True)

            Dim numClusters As Integer = 3
            Console.WriteLine(vbCrLf & "Setting numClusters to " & numClusters)

            Console.WriteLine(vbCrLf & "Starting clustering")
            Dim clustering() As Integer = Cluster(rawData, numClusters, 0)

            Console.WriteLine("Clustering complete" & vbCrLf)

            Console.WriteLine("Final clustering in internal format:" & vbCrLf)
            ShowVector(clustering, True)

            Console.WriteLine("Raw data by cluster:" & vbCrLf)
            ShowClustered(rawData, clustering, numClusters, 1)

            Console.WriteLine(vbCrLf & "End k-means++ clustering demo" & vbCrLf)
            Console.ReadLine()
        End Sub ' Main

        Public Shared Function Cluster(ByVal rawData()() As Double, ByVal numClusters As Integer, ByVal seed As Integer) As Integer()
            ' k-means++ clustering
            ' index of return is tuple ID, cell is cluster ID
            ' ex: [2 1 0 0 2 2] means tuple 0 is in cluster 2, tuple 1 in cluster 1,
            '  tuple 2 in cluster 0, tuple 3 in cluster 0, etc.
            Dim data()() As Double = Normalized(rawData) ' so large values don't dominate

            Dim changed As Boolean = True ' change in at least one cluster assignment?
            Dim success As Boolean = True ' all means computed? (no zero-count clusters)

            Dim means()() As Double = InitMeans(numClusters, data, seed) ' k-means++

            Dim clustering(data.Length - 1) As Integer ' all 0

            Dim maxCount As Integer = data.Length * 10 ' sanity check
            Dim ct As Integer = 0
            Do While changed = True AndAlso success = True AndAlso ct < maxCount ' technically this is Lloyd's algorithm
                changed = UpdateClustering(data, clustering, means) ' (re)assign tuples to clusters. no effect if fail
                success = UpdateMeans(data, clustering, means) ' compute new cluster means if possible. no effect if fail
                ct += 1 ' k-means typically converges very quickly
            Loop
            ' could check for 0-count clusters here
            Return clustering
        End Function

        Private Shared Function InitMeans(ByVal numClusters As Integer, ByVal data()() As Double, ByVal seed As Integer) As Double()()
            ' select k data items as initial means using k-means++ mechanism:
            ' pick one data item at random as first mean
            ' loop k-1 times (remaining means)
            '   compute dist^2 from each item to closest mean
            '   pick a data item w/ large dist^2 as next mean
            ' end loop

            Dim means()() As Double = MakeMatrix(numClusters, data(0).Length) ' result

            Dim used As New List(Of Integer)() ' track which items already means

            ' select one data item index at random
            Dim rnd As New Random(seed)
            Dim idx As Integer = rnd.Next(0, data.Length) ' [0, data.Length-1]
            Array.Copy(data(idx), means(0), data(idx).Length)
            used.Add(idx) ' we don't want to select this item again

            For k = 1 To numClusters - 1 ' each remaining mean
                Dim dSquared(data.Length - 1) As Double ' to closest mean
                Dim newMean As Integer = -1 ' index of data item to be a new mean
                For i = 0 To data.Length - 1 ' for each data item
                    ' if data item i is already a mean, skip the item
                    If used.Contains(i) = True Then ' not entirely necessary
                        Continue For
                    End If

                    ' compute distances from data[i] to each existing mean (to find closest)
                    Dim distances(k - 1) As Double ' we currently have k means
                    For j = 0 To k - 1
                        distances(j) = Distance(data(i), means(k)) ' could do dist^2 directly
                    Next j

                    ' now get the index of the closest mean
                    Dim m As Integer = MinIndex(distances)
                    ' save the associated distance-squared
                    dSquared(i) = distances(m) * distances(m)
                Next i

                ' pick one of the data items, using the squared distances
                ' this is a form of roulette wheel selection
                Dim p As Double = rnd.NextDouble()
                Dim sum As Double = 0.0 ' sum of distances-squared
                For i = 0 To dSquared.Length - 1
                    sum += dSquared(i)
                Next i
                Dim cumulative As Double = 0.0 ' cumulatiive probability

                Dim ii As Integer = 0 ' points into distancesSquared[]
                Dim sanity As Integer = 0 ' sanity count
                Do While sanity < data.Length * 2 ' 'stochastic acceptance'
                    cumulative += dSquared(ii) / sum
                    If cumulative >= p AndAlso used.Contains(ii) = False Then
                        newMean = ii ' the chosen index
                        used.Add(newMean) ' don't pick again
                        Exit Do
                    End If
                    ii += 1 ' next candidate
                    If ii >= dSquared.Length Then ' back to first item
                        ii = 0
                    End If
                    sanity += 1
                Loop
                ' check if newMean is still -1 . . . 

                ' save the data of the chosen index
                Array.Copy(data(newMean), means(k), data(newMean).Length)
            Next k ' k, each mean/cluster

            Return means

        End Function ' InitMeans

        Private Shared Function Normalized(ByVal rawData()() As Double) As Double()()
            ' normalize raw data by computing (x - mean) / stddev
            ' one alternative is min-max:
            ' v' = (v - min) / (max - min)

            ' make a copy of input data
            Dim result(rawData.Length - 1)() As Double
            For i = 0 To rawData.Length - 1
                result(i) = New Double(rawData(i).Length - 1) {}
                Array.Copy(rawData(i), result(i), rawData(i).Length)
            Next i

            For j = 0 To result(0).Length - 1 ' each col
                Dim colSum As Double = 0.0
                For i = 0 To result.Length - 1
                    colSum += result(i)(j)
                Next i
                Dim mean As Double = colSum / result.Length
                Dim sum As Double = 0.0
                For i = 0 To result.Length - 1
                    sum += (result(i)(j) - mean) * (result(i)(j) - mean)
                Next i
                Dim sd As Double = sum / result.Length
                For i = 0 To result.Length - 1
                    result(i)(j) = (result(i)(j) - mean) / sd
                Next i
            Next j
            Return result
        End Function

        Private Shared Function MakeMatrix(ByVal rows As Integer, ByVal cols As Integer) As Double()()
            ' convenience matrix allocator for Cluster()
            Dim result(rows - 1)() As Double
            For i = 0 To rows - 1
                result(i) = New Double(cols - 1) {}
            Next i
            Return result
        End Function

        Private Shared Function UpdateMeans(ByVal data()() As Double, ByVal clustering() As Integer, ByVal means()() As Double) As Boolean
            ' returns false if there is a cluster that has no tuples assigned to it
            ' parameter means[][] is really a ref parameter

            ' check existing cluster counts
            ' can omit this check if InitClustering and UpdateClustering
            ' both guarantee at least one tuple in each cluster (usually true)
            Dim numClusters As Integer = means.Length
            Dim clusterCounts(numClusters - 1) As Integer
            For i = 0 To data.Length - 1
                Dim cluster As Integer = clustering(i)
                clusterCounts(cluster) += 1
            Next i

            For k = 0 To numClusters - 1
                If clusterCounts(k) = 0 Then
                    Return False ' bad clustering. no change to means[][]
                End If
            Next k

            ' update, zero-out means so it can be used as scratch matrix 
            For k = 0 To means.Length - 1
                For j = 0 To means(k).Length - 1
                    means(k)(j) = 0.0
                Next j
            Next k

            For i = 0 To data.Length - 1
                Dim cluster As Integer = clustering(i)
                For j = 0 To data(i).Length - 1
                    means(cluster)(j) += data(i)(j) ' accumulate sum
                Next j
            Next i

            For k = 0 To means.Length - 1
                For j = 0 To means(k).Length - 1
                    means(k)(j) /= clusterCounts(k) ' danger of div by 0
                Next j
            Next k
            Return True
        End Function

        Private Shared Function UpdateClustering(ByVal data()() As Double, ByVal clustering() As Integer, ByVal means()() As Double) As Boolean
            ' (re)assign each tuple to a cluster (index of closest mean)
            ' returns false if no tuple assignments change OR
            ' if the reassignment would result in a clustering where
            ' one or more clusters have no tuples.

            Dim numClusters As Integer = means.Length
            Dim changed As Boolean = False

            Dim newClustering(clustering.Length - 1) As Integer ' proposed result
            Array.Copy(clustering, newClustering, clustering.Length)

            Dim distances(numClusters - 1) As Double ' from curr tuple to each mean

            For i = 0 To data.Length - 1 ' walk thru each tuple
                For k = 0 To numClusters - 1
                    distances(k) = Distance(data(i), means(k)) ' usually Euclidean
                Next k

                Dim newClusterID As Integer = MinIndex(distances) ' find closest mean ID
                'Console.WriteLine("new cluster Id = " + newClusterID);
                'Console.ReadLine();
                If newClusterID <> newClustering(i) Then
                    changed = True
                    newClustering(i) = newClusterID ' update
                End If
            Next i

            If changed = False Then
                Return False ' no change so bail and don't update clustering[][]
            End If

            ' check proposed clustering[] cluster counts
            Dim clusterCounts(numClusters - 1) As Integer
            For i = 0 To data.Length - 1
                Dim cluster As Integer = newClustering(i)
                clusterCounts(cluster) += 1
            Next i

            For k = 0 To numClusters - 1
                If clusterCounts(k) = 0 Then
                    Return False ' bad clustering. no change to clustering[][]
                End If
            Next k

            Array.Copy(newClustering, clustering, newClustering.Length) ' update
            Return True ' no empty clusters and at least one change
        End Function

        Private Shared Function Distance(ByVal tuple() As Double, ByVal mean() As Double) As Double
            ' Euclidean distance between two vectors for UpdateClustering()
            ' consider alternatives such as Manhattan distance
            Dim sumSquaredDiffs As Double = 0.0
            For j = 0 To tuple.Length - 1
                sumSquaredDiffs += Math.Pow((tuple(j) - mean(j)), 2)
            Next j
            Return Math.Sqrt(sumSquaredDiffs)
        End Function

        Private Shared Function MinIndex(ByVal distances() As Double) As Integer
            ' index of smallest value in array
            ' helper for UpdateClustering()
            Dim indexOfMin As Integer = 0
            Dim smallDist As Double = distances(0)
            For k = 0 To distances.Length - 1
                If distances(k) < smallDist Then
                    smallDist = distances(k)
                    indexOfMin = k
                End If
            Next k
            Return indexOfMin
        End Function

        Private Shared Sub ShowData(ByVal data()() As Double, ByVal decimals As Integer, ByVal indices As Boolean, ByVal newLine As Boolean)
            For i = 0 To data.Length - 1
                If indices Then
                    Console.Write(i.ToString().PadLeft(3) & " ")
                End If
                For j = 0 To data(i).Length - 1
                    If data(i)(j) >= 0.0 Then
                        Console.Write(" ")
                    End If
                    Console.Write(data(i)(j).ToString("F" & decimals) & " ")
                Next j
                Console.WriteLine("")
            Next i
            If newLine Then
                Console.WriteLine("")
            End If
        End Sub ' ShowData

        Private Shared Sub ShowVector(ByVal vector() As Integer, ByVal newLine As Boolean)
            For i = 0 To vector.Length - 1
                Console.Write(vector(i) & " ")
            Next i
            If newLine Then
                Console.WriteLine(vbCrLf)
            End If
        End Sub

        Private Shared Sub ShowVector(ByVal vector() As Double, ByVal decimals As Integer, ByVal newLine As Boolean)
            For i = 0 To vector.Length - 1
                Console.Write(vector(i).ToString("F" & decimals) & " ")
            Next i
            If newLine Then
                Console.WriteLine(vbCrLf)
            End If
        End Sub

        Private Shared Sub ShowClustered(ByVal data()() As Double, ByVal clustering() As Integer, ByVal numClusters As Integer, ByVal decimals As Integer)
            For k = 0 To numClusters - 1
                Console.WriteLine("===================")
                For i = 0 To data.Length - 1
                    Dim clusterID As Integer = clustering(i)
                    If clusterID <> k Then
                        Continue For
                    End If
                    Console.Write(i.ToString().PadLeft(3) & " ")
                    For j = 0 To data(i).Length - 1
                        If data(i)(j) >= 0.0 Then
                            Console.Write(" ")
                        End If
                        Console.Write(data(i)(j).ToString("F" & decimals) & " ")
                    Next j
                    Console.WriteLine("")
                Next i
                Console.WriteLine("===================")
            Next k ' k
        End Sub ' ShowClustered

    End Class ' Program class

End Namespace ' ns